/*-----------------------------------------------------------------------
  usbclock2 by Jack Botner

  main.c

  This project consists of the following hardware controlled by a
  host PC using an FTDI UB232R USB to serial interface.

  - 2x16 LCD display 
  - 2 pushbuttons 
  - beeper

  In this case the host sends time information, which is displayed
  on the LCD and used to sound the beeper on the hour. Button 1 is
  used to send a 'shutdown' message to the host program; button 2 
  changes time and date display from local to UTC and back.

  Processor: ATMega48
      Clock: Internal, 8 MHz (125 ns)

  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  20131218jb Changed use of button 2 from 'sound beeper' to
             'change date/timeformat'. The default is local
			 time. Pressing button 2 changes it it to UTC.
			 Pressing it again reverts to local time.
  20150608jb Implemented synchronization step at startup.
  jb20150822 - Reorganized checksum calculation, so I could turn off
               the high-order bit. This is to solve a communication
               failure after updating the FTDI driver. I guess the
               new driver doesn't believe '8 bits no parity' anymore.
-----------------------------------------------------------------------*/
#define F_CPU 8000000UL
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "lcd.h"

// Global variables

volatile unsigned char ucRxBuf[RX_BUF_SIZE];
volatile uint8_t ucRxBytes=0;
volatile unsigned char ucTxBuf[TX_BUF_SIZE];
volatile uint8_t ucTxBytes=0, ucTxIndex=0;
volatile uint8_t ucDiagCode=0;
volatile uint8_t ucDiagData=0;

volatile uint8_t ucDateTimeFormat=DATETIMELOCAL;
volatile uint8_t ucEventFlags=0;
volatile uint8_t uc100msCtr=0;
volatile uint16_t uc100msCtr2=ONE_MIN_OF_100MS;
volatile uint8_t ucSynchronized=0;

static const char *msg1 = "USB CLOCK 1.2";
static const char *msg2 = "- ATmega48 8MHz";

// Internal functions
static void diagnostic(void);
static void system_init(void);
static void process_id_msg(void);
static void process_time_msg(void);
static void process_date_msg(void);
static void send_shutdown_msg(void);
static void send_date_format_msg(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  system_init();                // initialise system
  
  lcd_init( LCD_DISP_ON_CURSOR );
  set_bit( PORTC, PC0 );        // turn on green diagnostic LED
  lcd_clrscr();
  lcd_gotoxy( 0, 0 );
  set_bit( PORTB, PB2 );		// backlight on
  lcd_puts( msg1 );

  while(1)
  {
    if ( test_bit( ucEventFlags, EVENT_RX_SYNCHRONIZE ) )
	{
      ucTxBuf[0] = 'a';				// synchronize response
      ucTxBytes = 1;
      ucTxIndex = 0;
      set_bit( UCSR0B, UDRIE0 );	// enable Tx interrupts
	  ucSynchronized = 1;
	  clear_bit( ucEventFlags, EVENT_RX_SYNCHRONIZE );
	}
	
    if ( test_bit( ucEventFlags, EVENT_RX_FRAME_ERROR ) )
    {
	  diagnostic();		// this stops until button 2 is pressed
	  ucRxBytes = 0;
	  clear_bit( ucEventFlags, EVENT_RX_FRAME_ERROR );
	}

    if ( test_bit( ucEventFlags, EVENT_RX_FRAME_COMPLETE ) )
	{
	  uc100msCtr2 = ONE_MIN_OF_100MS;	// reset the system timeout
	  set_bit( PORTB, PB2 );			// turn backlight on

	  if ( ucRxBuf[0] == 'T' )			// display time
	    process_time_msg();
	  else
	    if ( ucRxBuf[0] == 'D' )		// display date
	      process_date_msg();
	    else
	      if ( ucRxBuf[0] == 'S' )		// sound the beeper
		  {
		    set_bit( PORTC, PC4 );		// turn beeper on
			uc100msCtr = 3;				// for 300ms
	      }
		  else
	        if ( ucRxBuf[0] == 'M' )		// return your ID
			{
              lcd_init( LCD_DISP_ON_CURSOR );   // diagnostic test
		      process_id_msg();
			}
		    else
	          if ( ucRxBuf[0] == 'Z' )		// shutdown
				clear_bit( PORTB, PB2 );	// turn backlight off

	  ucRxBytes = 0;
	  clear_bit( ucEventFlags, EVENT_RX_FRAME_COMPLETE );
	}

    if ( test_bit( ucEventFlags, EVENT_INTERVAL_EXPIRED ) )
	{
	  clear_bit( PORTC, PC4 );		// turn beeper off
	  clear_bit( ucEventFlags, EVENT_INTERVAL_EXPIRED );
	}

    if ( test_bit( ucEventFlags, EVENT_SYSTEM_TIMEOUT ) )
	{
	  // The system timeout occurs after 1 minute of no traffic
	  // from the host program. Turn off the display.
	  clear_bit( PORTB, PB2 );		// turn backlight off
	  clear_bit( ucEventFlags, EVENT_SYSTEM_TIMEOUT );
	}

    if ( test_bit( ucEventFlags, EVENT_BUTTON1_RELEASED ) )
	{
	  // Send a 'shutdown' message to the host
	  send_shutdown_msg();
	  clear_bit( ucEventFlags, EVENT_BUTTON1_RELEASED );
	}

    if ( test_bit( ucEventFlags, EVENT_BUTTON2_RELEASED ) )
	{
	  // sound the beeper
  	  //set_bit( PORTC, PC4 );		// turn beeper on
  	  //uc100msCtr = 3;				// for 300ms

	  // 20131218jb change meaning of button 2 from sound beeper
	  // to change date/time format. Sends a message to the
	  // host to change the date/time format.
	  ucDateTimeFormat = ( ucDateTimeFormat == DATETIMELOCAL ) ? DATETIMEUTC : DATETIMELOCAL;
	  send_date_format_msg();
	  clear_bit( ucEventFlags, EVENT_BUTTON2_RELEASED );
	}

    _delay_loop_1( 3 );		// 1 us
  }
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (input) pushbutton 1
         PB1 (output) not used
         PB2 (output) LCD backlight
         PB3 (input) MOSI ISP
         PB4 (input) MISO ISP
         PB5 (input) SCK ISP
         PB6 (input) XTAL1
         PB7 (input) XTAL2

  PORTC: PC0 (output) green diagnostic LED
         PC1 (output) yellow diagnostic LED
         PC2 (output) LCD R/W
         PC3 (output) LCD E
         PC4 (output) Beeper
         PC5 (input) pushbutton 2
         PC6 (n/a) ISP !Reset

  PORTD: PD0 (input) serial RxD
         PD1 (output) serial TxD
         PD2 (output) not used
         PD3 (output) LCD RS
         PD4 (output) LCD DB4
         PD5 (output) LCD DB5
         PD6 (output) LCD DB6
         PD7 (output) LCD DB7
-----------------------------------------------------------------*/
static void system_init()
{
  // Make sure the clock prescaler is set to 1.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000000;       // set prescaler to 1
  
  // Set up the I/O ports directions, pull-ups
  DDRB  = 0b00000110;
  PORTB = 0b11000001;       // initialise PortB, 1 pullup

  DDRC  = 0b00011111;
  PORTC = 0b00100000;       // initialise PortC, 1 pullup

  DDRD  = 0b11111110;
  PORTD = 0b00000000;       // initialise PortD

  // Initialize timer0 for button debouncing
  TCNT0 = INIT_CTR0;
  TCCR0A = 0b00000000;       // timer/counter normal mode
  TCCR0B = 0b00000100;       // prescaler = 256 (turns timer0 on)
  set_bit( TIMSK0, TOIE0 );	// enable timer/counter0 overflow interrupt

  // Initialize timer1 for timing
  TCNT1 = INIT_CTR1;
  TCCR1A = 0b00000000;		// normal operation
  TCCR1B = 0b00000100;		// prescaler = 256 (turns timer1 on)
  TCCR1C = 0b00000000;		// normal operation
  set_bit( TIMSK1, TOIE1 );	// enable timer/counter1 overflow interrupt

  // Initialize the USART
  UBRR0L = 51;				// 9600 Baud (8 MHz clock) (spec table 20-1)
  UBRR0H = 0;
  UCSR0A = 0b00000000;		// double speed off, MPCM off
  UCSR0B = 0b10011000;		// Rx interrupts, rx/tx enable, 8 data bits
  UCSR0C = 0b00000110;		// async, no parity, 1 stop bit, 8 data bits

  sei();					// enable global interrupts
}

/*-----------------------------------------------------------------
  process_time_msg()

  The time message is a string that is shown on the 1st line.
-----------------------------------------------------------------*/
static void process_time_msg()
{
  char		cBuffer[17];
  uint8_t	nn;

  if ( ucRxBytes < 4 ) return;		// no data!
  nn = sizeof( cBuffer ) - 1;
  memset( cBuffer, ' ', nn );		// clean the buffer
  cBuffer[nn] = '\0';
  nn = ucRxBuf[1] - 3;				// compute data bytes
  memcpy( cBuffer, (const void *) ucRxBuf+2, nn );

  lcd_gotoxy( 0, 0 );
  lcd_puts( cBuffer );
}

/*-----------------------------------------------------------------
  process_date_msg()

  The time message is a string that is shown on the 2nd line.
-----------------------------------------------------------------*/
static void process_date_msg()
{
  char		cBuffer[17];
  uint8_t	nn;

  if ( ucRxBytes < 4 ) return;		// no data!
  nn = sizeof( cBuffer ) - 1;
  memset( cBuffer, ' ', nn );		// clean the buffer
  cBuffer[nn] = '\0';
  nn = ucRxBuf[1] - 3;				// compute data bytes
  memcpy( cBuffer, (const void *) ucRxBuf+2, nn );

  lcd_gotoxy( 0, 1 );
  lcd_puts( cBuffer );
}

/*-----------------------------------------------------------------
  process_id_msg()

  The id message is a string that is returned to the host.
-----------------------------------------------------------------*/
static void process_id_msg()
{
  uint8_t	ii, nn, ucCk;

  set_bit( PORTB, PB2 );		// turn backlight on
  set_bit( PORTC, PC4 );		// turn beeper on
  uc100msCtr = 3;				// for 300ms

  ucTxBuf[0] = 'N';				// id info
  strcpy( (char *) ucTxBuf+2, msg1 );
  strcat( (char *) ucTxBuf+2, msg2 );
  nn = strlen( (char *) ucTxBuf+2 ) + 2;
  ucTxBuf[1] = nn + 1;

  for ( ii=0, ucCk=0; ii<nn; ++ii )
    ucCk += ucTxBuf[ii];

  ucTxBuf[nn] = ucCk;
  ucTxBytes = ucTxBuf[1];
  ucTxIndex = 0;
  set_bit( UCSR0B, UDRIE0 );	// enable Tx interrupts
}

/*-----------------------------------------------------------------
  send_shutdown_msg()

  In response to pressing button 1, send a message to the host
  to shut down.
-----------------------------------------------------------------*/
static void send_shutdown_msg()
{
  ucTxBuf[0] = 'Z';				// shutdown cmd
  ucTxBuf[1] = 3;
  ucTxBuf[2] = ucTxBuf[0] + ucTxBuf[1];
  ucTxBytes = 3;
  ucTxIndex = 0;
  set_bit( UCSR0B, UDRIE0 );	// enable Tx interrupts
}

/*-----------------------------------------------------------------
  send_date_format_msg()

  In response to pressing button 2, send a message to the host
  to set the date/time format.
-----------------------------------------------------------------*/
static void send_date_format_msg()
{
  ucTxBuf[0] = ( ucDateTimeFormat == DATETIMELOCAL ) ? 'L' : 'U';
  ucTxBuf[1] = 3;
  ucTxBuf[2] = ucTxBuf[0] + ucTxBuf[1];
  ucTxBytes = 3;
  ucTxIndex = 0;
  set_bit( UCSR0B, UDRIE0 );	// enable Tx interrupts
}

/*-----------------------------------------------------------------
  diagnostic()

  Dump the Rx buffer
-----------------------------------------------------------------*/
static void diagnostic()
{
  static const char *msg = "Press Button 2";
  char cBuf[17], cTemp[4];
  uint8_t ii, nn;

  lcd_clrscr();
  lcd_gotoxy( 0, 0 );
  if ( ucDiagCode == 1 )
    sprintf( cBuf, "overflow %d", (unsigned) ucDiagData );
  else
    if ( ucDiagCode == 2 )
      sprintf( cBuf, "checksum %d", (unsigned) 	ucDiagData );
    else
    {
      nn = ( ucRxBytes > 8 ) ? 8 : ucRxBytes;
      for ( ii=0, cBuf[0]='\0'; ii<nn; ++ii )
      {
        sprintf( cTemp, "%02x", (unsigned) ucRxBuf[ii] );
	    strcat( cBuf, cTemp );
      }
	}
	  
  lcd_puts( cBuf );
  lcd_gotoxy( 0, 1 );
  lcd_puts( msg );

  while ( !test_bit( ucEventFlags, EVENT_BUTTON2_RELEASED ) );
  clear_bit( ucEventFlags, EVENT_BUTTON2_RELEASED );
  clear_bit( ucEventFlags, EVENT_SYSTEM_TIMEOUT );
  ucDiagCode = 0;
  ucRxBytes = 0;
}
